
class FunctionWrapper:
    def __init__(self, inner_func):
        self.inner_func = inner_func

    def before(self, *args, **kwargs):
        pass

    def after(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        self.before(*args, **kwargs)
        val = self.inner_func(*args, **kwargs)
        self.after(val, *args, **kwargs)
        return val

    @classmethod
    def method_hook(cls, target, name):
        target_func = getattr(target, name)
        wrapped = cls(target_func)

        def func(self, *args, **kwargs):
            return wrapped(self, *args, **kwargs)
        setattr(target, name, func)
        return wrapped

    @classmethod
    def function_hook(cls, func):
        return cls(func)




import numpy as np
import typing
from math import ceil, floor
from .networkx_stubs import TrackedList as TList
from .networkx_stubs import get_last_trace_id, add_to_ops_accum
from .networkx_stubs import TraceDepRegistry as tdr
from .networkx_stubs import TraceRegistry as tr

# Tile size is 4MB, 1M ints
TILE_SIZE = int(2 * 1024 * 1024 / 4)


class TArrayTraceElement:
    """
    A TArray element for book keeping.
    The trace_id_list contains all the trace ids this element depends on.
    TODO: need to find a way to dispatch these functions in one line. One
    proposal is that I could use the inspect library to look at the function
    vars, and then annotate them with my own subroutine.
    """

    def _update_nops_accum(self):
        add_to_ops_accum(
            len(self.value) if isinstance(
                self.value.size, (np.ndarray, TArray)
            ) else 1
        )

    def _merge_trace_id_lists(self, b):
        default_trace_list = self.trace_id_list
        if isinstance(b, TArrayTraceElement):
            default_trace_list += b.trace_id_list
        return default_trace_list

    def _merge_value(self, b, fn: typing.Callable):
        default_value = self.value
        if isinstance(b, TArrayTraceElement):
            default_value = fn(self.value, b)
        return default_value

    def __init__(self, value, trace_id_list: typing.List) -> None:
        self.value = value
        self.trace_id_list = trace_id_list

    def __iter__(self):
        """
        If we are iterating on an TArrayTraceElement, we can bulk-load
        the actual array.
        :return:
        """
        return self.value.__iter__()

    def __le__(self, other):
        return self.value <= other

    def __eq__(self, other):
        return self.value == other

    def __gt__(self, other):
        return self.value > other

    def __lt__(self, other):
        return self.value < other

    def __ge__(self, other):
        return self.value >= other

    def __mul__(self, other):
        self._update_nops_accum()
        return TArrayTraceElement(
            value=self.value * other.value if isinstance(
                other, TArrayTraceElement
            ) else self.value * other,
            trace_id_list=self._merge_trace_id_lists(other)
        )

    def __rmul__(self, other):
        return self.__mul__(other)

    def __imul__(self, other):
        return self.__mul__(other)

    def __add__(self, other):
        self._update_nops_accum()
        return TArrayTraceElement(
            value=self.value + other.value if isinstance(
                other, TArrayTraceElement
            ) else self.value + other,
            trace_id_list=self._merge_trace_id_lists(other)
        )

    def __radd__(self, other):
        return self.__add__(other)

    def __iadd__(self, other):
        return self.__add__(other)

    def __truediv__(self, other):
        self._update_nops_accum()
        return TArrayTraceElement(
            value=self.value / other.value if isinstance(
                other, TArrayTraceElement
            ) else self.value / other,
            trace_id_list=self._merge_trace_id_lists(other)
        )

    def __itruediv__(self, other):
        return self.__truediv__(other)

    def __rtruediv__(self, other):
        self._update_nops_accum()
        return TArrayTraceElement(
            value=other.value / self.value if isinstance(
                other, TArrayTraceElement
            ) else other / self.value,
            trace_id_list=self._merge_trace_id_lists(other)
        )

    def __sub__(self, other):
        self._update_nops_accum()
        return TArrayTraceElement(
            value=self.value - other.value if isinstance(
                other, TArrayTraceElement
            ) else self.value - other,
            trace_id_list=self._merge_trace_id_lists(other)
        )

    def __isub__(self, other):
        return self.__sub__(other)

    def __rsub__(self, other):
        self._update_nops_accum()
        return TArrayTraceElement(
            value=other.value - self.value if isinstance(
                other, TArrayTraceElement
            ) else other - self.value,
            trace_id_list=self._merge_trace_id_lists(other)
        )

    def sum(self):
        self._update_nops_accum()
        return TArrayTraceElement(
            value=self.value.sum(),
            trace_id_list=self.trace_id_list
        )

    def dot(self, other):
        self._update_nops_accum()
        return TArrayTraceElement(
            value=self.value.dot(other),
            trace_id_list=self.trace_id_list
        )


class NpSet(set):
    """
    Numpy ndarray is not hashable. However a tuple of a flattened numpy
    array is.
    This function wraps a numpy ndarray with a tuple wrapper.
    """

    def add(self, element) -> None:
        if isinstance(element, TArrayTraceElement):
            element = element.value
        elif isinstance(element, TArray):
            element = TArray.view(np.ndarray)
        super().add(tuple(element.flatten()))

    def __contains__(self, item):
        if isinstance(item, TArray):
            item = item.view(np.ndarray)
        return super(NpSet, self).__contains__(tuple(item.flatten()))


ProfiledArraySet: typing.Set = NpSet()


def _profiled_check(items: typing.List):
    """
    Disable profiled check for now as this seems to incur unnecessary tests...
    :param items:
    :return:
    """
    # for k in items:
    #     if k not in ProfiledArraySet:
    #         ProfiledArraySet.add(k)


class TArray(np.ndarray):
    """
    We don't need an initializer here; this class is a thin wrapper
    to hijack the runtime for @. Otherwise monkey patching would fail
    due to @ being a builtin.

    TODO: there might be a better way to call subroutines defined in the
    super class. Should be a way to just decorate all of these.

    TODO: we may need to also subclass TList. For now I'm doing the
    bookkeeping within the _init_book_keeping function...
    """
    def _init_book_keeping(self) -> None:
        """
        TODO: remove the is_host_init. Do it at init time for TList.
        :return:
        """
        if '_tlist' not in vars(self):
            self._tlist = TList([0] * self.size, is_host_init=True)
            self._tlist._is_host_init = False
            self._is_iter = False
            self._iter_id = -1

    def _update_blas_3_ops(self, a, b):
        if isinstance(
            a, (TArray, np.ndarray)
        ) and isinstance(
            b, (TArray, np.ndarray)
        ):
            a_n_rows, a_n_cols = a.shape
            _, b_n_cols = b.shape
            add_to_ops_accum(
                a_n_rows * a_n_cols * b_n_cols * 2
            )

    def _update_blas_1_map_ops(self, other):
        if isinstance(other, (TArray, np.ndarray)):
            add_to_ops_accum(len(other))

    def _update_blas_1_map_reduce_ops(self, other):
        if isinstance(other, (TArray, np.ndarray)):
            add_to_ops_accum(len(other) * 2)

    def _blas_3_stub(self, *args):
        global ProfiledArraySet
        other = args[0]
        self._init_book_keeping()
        self._update_blas_3_ops(self, other)

        if isinstance(other, (TArray, np.ndarray)) and \
                other not in ProfiledArraySet:
            if other.size > TILE_SIZE or self.size > TILE_SIZE:
                # We need to do tiling here for both reads and writes
                def _get_row_col_steps(a_shape: typing.Tuple,
                                       b_shape: typing.Tuple):
                    """
                    Once getting a shape, this function should return the
                    number of steps required to step through both dimensions
                    based on the given tile size.
                    :param shape:
                    :return: m_steps, k_steps, n_steps, tile_size on the
                    reduce dim
                    """
                    m, _k = a_shape
                    __k, n = b_shape
                    if _k != __k:
                        print("Warning: inner product dimensions don't match!")

                    k = _k

                    _step_size = ceil(TILE_SIZE / float(k))
                    _m_steps = ceil(m / _step_size)
                    _n_steps = ceil(n / _step_size)
                    _k_steps = ceil(
                        float(k) / TILE_SIZE
                    ) if _step_size == 1 else 1
                    _m_residual_step_size = m % _step_size
                    return _m_steps, _k_steps, _n_steps, _step_size, \
                        m % _step_size, n % _step_size

                m_steps, k_steps, n_steps, step_size, \
                m_residual_step_size, n_residual_step_size \
                    = _get_row_col_steps(self.shape, other.shape)

                # Each tiled multiplication is within a map scope; hence
                # there's no dependency here.

                par_dep = get_last_trace_id()
                for i_m in range(m_steps):
                    for i_n in range(n_steps):

                        # Issue all the reads for doing this inner matmul
                        for i_k in range(k_steps):
                            # Load two read tiles of size TILE_SIZE
                            def _get_read_trace_id():
                                _tmp = TList(
                                    [0] * TILE_SIZE, is_host_init=True
                                )
                                _, trace_id = _tmp.__getitem__(
                                    slice(None, None, None),
                                    deps=[par_dep]
                                )

                                return trace_id

                            read_deps = [
                                _get_read_trace_id() for i in range(2)
                            ]

                        # Issue tiled writes
                        m_curr_step_size = step_size \
                            if i_m < m_steps - 1 else m_residual_step_size
                        n_curr_step_size = step_size \
                            if i_n < n_steps - 1 else n_residual_step_size
                        write_size = m_curr_step_size * n_curr_step_size
                        write_steps = ceil(write_size / float(TILE_SIZE))

                        for i_w in range(write_steps):
                            _tmp = TList(
                                [0] * TILE_SIZE,
                                is_host_init=True
                            )
                            _tmp.__setitem__(
                                slice(None, None, None),
                                slice(None, None, None),
                                deps=read_deps
                            )

            else:
                tmp = TList(
                    [0] * other.size, is_host_init=True
                )
                _ = tmp.__getitem__(
                    slice(None, None, None),
                    deps=[get_last_trace_id()]
                )

            _profiled_check([self.T, other.T])

    def _blas_1_stub(self, *args):
        global ProfiledArraySet
        self._init_book_keeping()

        add_to_ops_accum(self.size)

        if args:
            other = args[0]
            if isinstance(other, (TArray, np.ndarray)) and \
                    other not in ProfiledArraySet:
                tmp = TList(
                    [0] * other.size, is_host_init=True
                )
                _ = tmp.__getitem__(
                    slice(None, None, None),
                    deps=[get_last_trace_id()]
                )

                _profiled_check([self.T, other.T])

    @staticmethod
    def _extract_val_trace_id_list(a):
        if isinstance(a, TArrayTraceElement):
            return a.value, a.trace_id_list
        else:
            return a, []

    def __iter__(self, *args, **kwargs):
        self._init_book_keeping()
        if not self._is_iter:
            self._is_iter = True
            _, self._iter_id = self._tlist.__getitem__(
                slice(None, None, None),
                deps=[get_last_trace_id()]
            )
        return super().__iter__(*args, **kwargs)

    def __setitem__(self, *args, **kwargs):
        """
        This function is to capture dependencies. For a WA_, it would take
        the form of:
        A = foo.view(TArray)
        A[idx, idx_trace_id_0] = c, c_trace_id # Write
        _, _ = A[idx, idx_trace_id_0] # Read
        or
        A[idx, idx_trace_id_0] = _, _ # Write
        :param args:
        :param kwargs:
        :return:
        """

        self._init_book_keeping()
        k, v = args

        key, key_trace_list = self._extract_val_trace_id_list(k)
        val, val_trace_list = self._extract_val_trace_id_list(v)

        # TODO: for now we assume that writes don't generate dependency.
        # This is definitely wrong but seems to be good enough for many of
        # our apps.
        # TODO: in order to get the write through, I'm flipping the
        # host_init attribute of self. Not a neat way to do and I will need
        # to fix it in the future release. The current patch adds a barrier
        # around writes.
        self._tlist.__setitem__(
            key, val, deps=key_trace_list + val_trace_list
        )
        super().__setitem__(key, val)

    def __getitem__(self, *args, **kwargs):
        # Seems that when calling the view function, the *args are in fact
        # tuples. We only want to init getitem and setitem when the *args
        # are either int or slices.
        # TODO: the instances checks are here to guard against weird
        # accesses thrown by the view function. Right now I'm patching it
        # with these guards. Later we need to understand how these view
        # functions work... The debugger also cannot catch these issues either.
        v = args[0]
        if not isinstance(v, tuple):
            self._init_book_keeping()
            if isinstance(
                    v, (int, np.int, np.int64)
            ) and v >= 0:
                if self._is_iter:
                    # print("getting element from iter, element = {}".format(v))
                    trace_id = self._iter_id
                else:
                    # print("getting single element, element = {}".format(v))
                    _, trace_id = self._tlist.__getitem__(v, deps=[])
                result = TArrayTraceElement(
                    super(TArray, self).__getitem__(v),
                    trace_id_list=[trace_id]
                )
            elif isinstance(v, slice):
                # print("getting slice, slice = {}".format(v))
                slice_start_val, slice_start_trace_id_list = \
                    self._extract_val_trace_id_list(v.start)
                slice_stop_val, slice_stop_trace_id_list = \
                    self._extract_val_trace_id_list(v.stop)
                v = slice(slice_start_val, slice_stop_val)
                _, trace_id = self._tlist.__getitem__(
                    v,
                    deps=slice_start_trace_id_list + slice_stop_trace_id_list
                )
                result = TArrayTraceElement(
                    super(TArray, self).__getitem__(v),
                    trace_id_list=[trace_id]
                )

            elif isinstance(v, TArrayTraceElement):
                # print("getting a TATElement from TArray, value = {}".format(v))
                if self._is_iter:
                    trace_id = self._iter_id
                else:
                    _, trace_id = self._tlist.__getitem__(
                        v.value, deps=v.trace_id_list
                    )
                result = TArrayTraceElement(
                    super(TArray, self).__getitem__(v.value),
                    trace_id_list=[trace_id]
                )
            else:
                print("Wrong format {}...".format(v))
                result = super(TArray, self).__getitem__(*args, **kwargs)
        elif isinstance(v, tuple):
                # print(
                    # "initing in the view function..., value = {}".format(v)
                # )
                result = super(TArray, self).__getitem__(*args, **kwargs)
        else:
            print("Wrong format {}...".format(v))
            result = super(TArray, self).__getitem__(*args, **kwargs)

        return result
        
    def dot(self, *args, **kwargs):
        # print("__dot__")
        self._blas_1_stub(*args)
        self._update_blas_1_map_ops(self) # Add one more here for reduction
        return super(TArray, self).dot(*args, **kwargs)

    def __add__(self, *args, **kwargs):
        # print("add")
        self._blas_1_stub(*args)
        return super(TArray, self).__add__(*args, **kwargs)

    def __iadd__(self, *args, **kwargs):
        # print("iadd")
        self._blas_1_stub(*args)
        return super(TArray, self).__iadd__(*args, **kwargs)

    def __radd__(self, *args, **kwargs):
        # print("radd")
        self._blas_1_stub(*args)
        return super(TArray, self).__radd__(*args, **kwargs)

    def sum(self, *args, **kwargs):
        # print("sum")
        # stubbing is done already...
        # self._blas_1_stub(*args)

        # ops counting is not
        self._update_blas_1_map_ops(self)
        return super(TArray, self).sum(*args, **kwargs)

    def __matmul__(self, *args, **kwargs):
        self._blas_3_stub(*args)
        return super(TArray, self).__matmul__(*args, **kwargs)

    def __rmatmul__(self, *args, **kwargs):
        # print("__rmatmul__")
        self._blas_1_stub(*args)
        return super(TArray, self).__rmatmul__(*args, **kwargs)

    def __mul__(self, *args, **kwargs):
        # print("__mul__")
        self._blas_1_stub(*args)
        return super(TArray, self).__mul__(*args, **kwargs)

    def __rmul__(self, *args, **kwargs):
        # print("__rmul__")
        self._blas_1_stub(*args)
        return super(TArray, self).__rmul__(*args, **kwargs)

    def __imul__(self, *args, **kwargs):
        # print("__imul__")
        self._blas_1_stub(*args)
        return super(TArray, self).__imul__(*args, **kwargs)


def array(fn, *args, **kwargs):
    result = fn(*args, **kwargs)
    return result.view(TArray)




from .networkx_stubs import TrackedList as TList
from scipy.sparse import _sparsetools
from functools import partial

from torch import sparse
# For debugging
from .networkx_stubs import TraceRegistry as tr
from .networkx_stubs import TraceDepRegistry as tdr
from .networkx_stubs import get_last_trace_id, add_to_ops_accum

_n_bytes = 4


def csr_matmat_pass1(fn, *args, **kwargs):
    """
    We are assuming that this pass would keep all the necessary data in
    and can fuse with csr_matmat_pass2...
    We are also assuming that the W is pipelined with R.

    Pseudo code
    for i in range(n_row):
        row_nnz = 0
        for jj in range(Ap[i], Ap[i+1]): // sparse read
            j = Aj[jj] // sparse read
            for kk in range(Bp[j], Bp[j+1]): // sparse read
                k = Bj[kk] // dense
                if not mask[k] == i:
                    mask[k] = i
                    row_nnz += 1

        nnz = nnz + row_nnz
        Cp[i+1] = nnz

    :param fn:
    :param args:
    :param kwargs:
    :return:
    """

    print("in csr_matmat_pass1")
    n_row, n_col, *tail = args
    _, Aj_raw, Bp_raw, Bj_raw, _ = tail
    n = len(Bp_raw)
    P = float(len(Aj_raw)) / float(n_row * n)
    Q = float(len(Bj_raw)) / float(n_col * n)
    # We are using the assumption that the probability of nnzs are independent.
    add_to_ops_accum(
        n_row * n_col * n * P * Q * 2
    )

    Ap, Aj, Bp, Bj, Cp = [TList(x, is_host_init=True) for x in tail]

    Ap_last_ptr, Ap_last_ptr_trace_id = Ap.__getitem__(
        0, deps=[get_last_trace_id()]
    )

    for i in range(n_row):
        Ap_curr_ptr, Ap_curr_ptr_trace_id = Ap.__getitem__(
            i+1, deps=[get_last_trace_id()]
        )
        js, js_trace_id = Aj.__getitem__(
            slice(Ap_last_ptr, Ap_curr_ptr),
            deps=[Ap_last_ptr_trace_id, Ap_curr_ptr_trace_id]
        )
        for j in js:
            kk_start, kk_start_trace_id = Bp.__getitem__(
                j, deps=[js_trace_id]
            )
            kk_end, kk_end_trace_id = Bp.__getitem__(
                j+1, deps=[js_trace_id]
            )
            _ = Bj.__getitem__(
                slice(kk_start, kk_end),
                deps=[kk_start_trace_id, kk_end_trace_id]
            )

        Ap_last_ptr = Ap_curr_ptr
        Ap_last_ptr_trace_id = Ap_curr_ptr_trace_id

        Cp[i+1] = 0

    return fn(*args, **kwargs)


def csr_matmat_pass2(fn, *args, **kwargs):
    """
    For now, we are not profiling this function.
    It seems that with compiler fusion, we should be able to
    execute this with pass 1 in the same space.
    """
    # Pseudo code:
    # next = [-1] * n_col
    # sums = [0] * n_col
    # nnz = 0
    # Cp[0] = 0
    #
    # for i in range(n_row):
    #     head = -2
    #     length = 0
    #     jj_start = Ap[i]
    #     jj_end = Ap[i+1]
    #     for jj in range(jj_start, jj_end):
    #         j = Aj[jj]
    #         v = Ax[jj]
    #         kk_start = Bp[j]
    #         kk_end = Bp[j+1]
    #         for kk in range(kk_start, kk_end):
    #             k = Bj[kk]
    #             sums[k] += v * Bx[kk]
    #
    #             if next[k] == -1:
    #                 next[k] = head
    #                 head = k
    #                 length += 1
    #
    #     for jj in range(length):
    #         if not sums[head] == 0:
    #             Cj[nnz] = head
    #             Cx[nnz] = sums[head]
    #             nnz += 1
    #
    #         temp = head
    #         head = next[head]
    #         next[temp] = -1
    #         sums[temp] = 0
    #
    #         Cp[i+1] = nnz

    return fn(*args, **kwargs)


def csr_matvec(fn, *args, **kwargs):
    """
    Pseudo code
    for i in range(n_row):
        sum = Yx[i]
        for jj in range(Ap[i], Ap[i+1]):
            sum += Ax[jj] * Xx[Aj[jj]]
        Yx[i] = sum
    """
    print("in csr_matvec")

    n_row, n_col, *tail = args
    _, _, Ax_raw, _, _ = tail
    nnz = len(Ax_raw)
    add_to_ops_accum(nnz * 2)

    Ap, Aj, Ax, Xx, Yx = [TList(x, is_host_init=True) for x in tail]

    Yx_n_row, Yx_n_trace_id = Yx[0:n_row]
    Ap_last_ptr, Ap_last_ptr_trace_id = Ap.__getitem__(
        0, deps=[get_last_trace_id()]
    )
    for i, e in enumerate(Yx_n_row):
        Ap_next_ptr, Ap_next_ptr_trace_id = Ap.__getitem__(
            i+1, deps=[get_last_trace_id()]
        )
        Ax.__getitem__(
            slice(Ap_last_ptr, Ap_next_ptr),
            deps=[Ap_last_ptr_trace_id, Ap_next_ptr_trace_id]
        )
        # _ = Ax[Ap_last_ptr:Ap_next_ptr]
        Aj_sub, Aj_sub_trace_id = Aj.__getitem__(
            slice(Ap_last_ptr, Ap_next_ptr),
            deps=[Ap_last_ptr_trace_id, Ap_next_ptr_trace_id]
        )
        # Aj_sub = Aj[Ap_last_ptr:Ap_next_ptr]
        for j_Aj in Aj_sub:
            # _ = Xx[j_Aj]
            Xx.__getitem__(
                j_Aj, deps=[Aj_sub_trace_id]
            )
        Ap_last_ptr = Ap_next_ptr
        Ap_last_ptr_trace_id = Ap_next_ptr_trace_id

    return fn(*args, **kwargs)


def csr_matvecs(fn, *args, **kwargs):
    """
    Pseudo code
    for i in range(n_row):
        Y_ptr = Yx + n_vecs * i             # 2 * n_row ops
        for jj in range(Ap[i], Ap[i+1):     # nnz
            j = Aj[jj]
            a = Ax[jj]
            x = Xx + n_vecs * j             # 2 * nnz ops
            axpy(n_vecs, a, x, y)           # n_vecs * 2 * nnz ops
    :param fn:
    :param args:
    :param kwargs:
    :return:
    """
    print("in csr_matvecs")
    n_row, _, n_vecs, *tail = args
    Ap_raw = tail[0]
    nnz = len(Ap_raw)
    add_to_ops_accum(
        n_row * 2 + nnz * (2 + n_vecs * 2)
    )
    Ap, Aj, Ax, Xx, Yx = [
        TList(x, is_host_init=True) for x in tail
    ]
    Ap_last_ptr, Ap_last_ptr_trace_id = Ap.__getitem__(
        0, deps=[get_last_trace_id()]
    )

    for i in range(n_row):
        y_ptr = n_vecs * i
        _, Yx_trace_id = Yx.__getitem__(
            slice(y_ptr, y_ptr+n_vecs),
            deps=[get_last_trace_id()]
        )
        Ap_new_ptr, Ap_new_ptr_trace_id = Ap.__getitem__(
            i+1, deps=[get_last_trace_id()]
        )

        # Aj_jj_sublist = [Ap_last_ptr:Ap_new_ptr]
        # Ax_jj_sublist = Ax[Ap_last_ptr:Ap_new_ptr]

        Aj_jj_sublist, Aj_jj_sublist_trace_id = Aj.__getitem__(
            slice(Ap_last_ptr, Ap_new_ptr),
            deps=[Ap_last_ptr_trace_id, Ap_new_ptr_trace_id]
        )
        Ax_jj_sublist, Ax_jj_sublist_trace_id = Ax.__getitem__(
            slice(Ap_last_ptr, Ap_new_ptr),
            deps=[Ap_last_ptr_trace_id, Ap_new_ptr_trace_id]
        )

        for j, a in zip(Aj_jj_sublist, Ax_jj_sublist):
            x_ptr = n_vecs * j
            # _ = Xx[x_ptr:x_ptr+n_vecs]
            _ = Xx.__getitem__(
                slice(x_ptr, x_ptr+n_vecs),
                deps=[Aj_jj_sublist_trace_id, Ax_jj_sublist_trace_id]
            )

            # We assume that axpy is DRAM bound.
            # Hence once the data is ready, we don't need to simulate the
            # compute.

        Ap_last_ptr = Ap_new_ptr
        Ap_last_ptr_trace_id = Ap_new_ptr_trace_id

    return fn(*args, **kwargs)


def csc_matvec(fn, *args, **kwargs):
    print("WARNING: csc matvec is not yet supported; please use csr matvec.")
    return fn(*args, **kwargs)


def csc_matvecs(fn, *args, **kwargs):
    print("WARNING: csc matvecs is not yet supported; please use csc matvecs.")
    return fn(*args, **kwargs)


def stub():
    _sparsetools.csr_matmat_pass1 = partial(
        csr_matmat_pass1, _sparsetools.csr_matmat_pass1
    )
    _sparsetools.csr_matmat_pass2 = partial(
        csr_matmat_pass2, _sparsetools.csr_matmat_pass2
    )
    _sparsetools.csr_matvec = partial(
        csr_matvec, _sparsetools.csr_matvec
    )
    _sparsetools.csr_matvecs = partial(
        csr_matvecs, _sparsetools.csr_matvecs
    )
    _sparsetools.csc_matvec = partial(
        csc_matvec, _sparsetools.csc_matvec
    )
    _sparsetools.csc_matvecs = partial(
        csc_matvecs, _sparsetools.csc_matvecs
    )



import networkx
import itertools
import typing
import random
import sys
import weakref


class TraceElement(typing.NamedTuple):
    memory_space: int
    offset: int
    size: int
    type: str
    iterator_id: int = -1


class TraceDepElement(typing.NamedTuple):
    dep_on: typing.List[int]
    edges_out: typing.List[int]


TraceElement.READ = "R"
TraceElement.WRITE = "W"


class MemorySpace(typing.NamedTuple):
    memory_space: int
    size: int  # in number of bytes
    element_size: int  # in number of bytes


TraceDepRegistry: typing.List[TraceDepElement] = []


class RegistryList(list):

    def append(
            self, object: typing.Callable[[], TraceElement],
            deps: typing.List[int] = None,
            out_edges: typing.List[int] = None
    ) -> typing.Optional[int]:
        curr_id = len(TraceRegistry)
        deps = list(filter(lambda x: x >= 0, deps)) if deps else []
        out_edges = out_edges if out_edges else []
        TraceDepRegistry.append(TraceDepElement(deps, out_edges))
        for dep_id in deps:
            if dep_id >= 0: # Guard against the first access
                TraceDepRegistry[dep_id].edges_out.append(curr_id)
        super().append(object)

        return curr_id


MemorySpaceCounter = itertools.count()
IteratorCounter = itertools.count()

# TraceRegistry: typing.List[typing.Callable[[], TraceElement]] = []
TraceRegistry = RegistryList()
MemorySpaceRegistry: typing.List[typing.Callable[[], MemorySpace]] = []
OpsAccum: int = 0


def add_to_ops_accum(nops: int):
    global OpsAccum
    OpsAccum += nops

def get_ops_accum():
    global OpsAccum
    return OpsAccum


def get_last_trace_id():
    """
    This function is used for getting the last trace id.
    Generally speaking, we need to obtain the dependencies between traces.
    If the program exits a stubbed function, the next function,
    when starting off, needs to depend the first trace on the last trace
    recorded.
    :return:
    """
    return len(TraceRegistry) - 1


def dumptrace(tracefile=sys.stdout):
    global OpsAccum
    print("Number of Ops: {}".format(int(OpsAccum)))
    print("Traces:", len(TraceRegistry))
    for trace_id, (trace, trace_dep) in enumerate(
            zip(TraceRegistry[:], TraceDepRegistry[:])):
        try:
            print(*trace(), file=tracefile, sep="\t")
            # print(trace_id, file=tracefile, sep="\t", end="\t")
            # print(*trace(), file=tracefile, sep="\t", end="\t")
            # print(*trace_dep, file=tracefile, sep="\t")

        except TypeError:
            # Networkx has a bad habit of catching TypeErrors
            pass


def dumpmem(spacefile=sys.stdout):
    print("Spaces:", len(MemorySpaceRegistry))
    for space in MemorySpaceRegistry[:]:
        print(*space(), file=spacefile, sep="\t")


def reset():
    TraceRegistry.clear()


class TrackedIterator(typing.Iterator):
    def __init__(self, inner_iter, container: 'TrackedContainer',
                 trace_size=1):
        self.container = container
        self.inner_iter = enumerate(inner_iter)
        self.trace_size = trace_size
        self.iterator_id = next(IteratorCounter)

    def __next__(self):
        count, value = next(self.inner_iter)
        if count % self.trace_size == 0:
            TraceRegistry.append(lambda: self.container.getbase()._replace(
                offset=count * self.container.element_size,
                size=max(self.trace_size * self.container.element_size,
                         self.container.getsize()),
                type=TraceElement.READ,
                iterator_id=self.iterator_id
            ))
        return value


class TrackedContainer:
    def __init_subclass__(cls, is_sparse=False):
        cls.is_sparse = is_sparse

    parent: 'TrackedContainer'
    key: typing.Hashable
    max_len: int
    max_element_size_registry: typing.List[
        typing.Union[typing.Callable[[], int]]]

    # Overhead
    base_size: int = 4

    # feature size
    feature_size: int = 4

    key_to_loc_map: typing.MutableMapping[typing.Hashable, int]

    def __init__(self, *args, is_host_init=False, max_size=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.parent = None
        self.key = None
        self._memory_space_id = None
        self._is_host_init = is_host_init
        self.max_len = len(self)
        self.max_element_size_registry = []

        self._max_size = max_size

        self.key_to_loc_map = {}

    def _parse_slice(self, key: slice):
        start = key.start if key.start else 0
        stop = key.stop if key.stop else len(self)
        length = stop - start
        return start, stop, length

    def __setitem__(self, key, value, deps: typing.List = None):
        if isinstance(value, TrackedContainer):
            value.parent = self
            value.key = key
            self.max_element_size_registry.append(
                weakref.ref(value, self.max_size_callback))
        else:
            self.max_element_size_registry.append(self.feature_size)

        self.max_len = max(self.max_len, len(self))

        # If is not a host init, we should assume that the data is prepared
        # in the DRAM and the accelerator can just read from it.
        trace_id = get_last_trace_id()

        if isinstance(value, slice):
            start, _, length = self._parse_slice(value)
            trace_id = TraceRegistry.append(
                lambda: self.getloc(start)._replace(
                    type=TraceElement.WRITE,
                    size=self.feature_size * length
                ),
                deps=deps
            )
        elif not self._is_host_init:
            trace_id = TraceRegistry.append(
                lambda: self.getloc(key)._replace(type=TraceElement.WRITE),
                deps=deps
            )

            super().__setitem__(key, value)

        return trace_id

    def __getitem__(
            self, key, deps: typing.List[int] = [],
            out_edges: typing.List[int] = []
    ):
        val = super().__getitem__(key)

        if isinstance(key, slice):
            # Issue a dense read for slices
            start, _, length = self._parse_slice(key)
            trace_id = TraceRegistry.append(
                lambda: self.getloc(start)._replace(
                    type=TraceElement.READ,
                    size=self.feature_size * length
                ),
                deps=deps,
                out_edges=out_edges
            )
        elif not isinstance(val, TrackedContainer):
            # Reading into another container can be bypassed via smart address calculation
            trace_id = TraceRegistry.append(
                lambda: self.getloc(key)._replace(
                    type=TraceElement.READ
                ),
                deps=deps,
                out_edges=out_edges
            )
        return val, trace_id

    def __delitem__(self, key):
        TraceRegistry.append(lambda: self.getloc(key)._replace(type="DELETE"))
        super().__delitem__(self, key)

    def __iter__(self):
        return TrackedIterator(super(TrackedContainer, self).__iter__(), self)

    @property
    def memory_space_id(self):
        if self.parent:
            return self.parent.memory_space_id
        if self._memory_space_id is None:
            self._memory_space_id = next(MemorySpaceCounter)
            MemorySpaceRegistry.append(
                lambda: MemorySpace(self.memory_space_id, self.getsize(),
                                    self.element_size)
            )
        return self._memory_space_id

    def getsize(self) -> int:
        return self.element_size * self.max_len + self.base_size

    @property
    def element_size(self) -> int:
        if self._max_size is not None:
            return self._max_size
        max_size = 0
        for element in self.max_element_size_registry:
            if isinstance(element, int):
                max_size = max(element, max_size)
            if isinstance(element, typing.Callable):
                fetched = element()
                if fetched is not None:
                    max_size = max(fetched.getsize(), max_size)

        self._max_size = max_size
        return self._max_size

    def max_size_callback(self, subcontainer):
        self.max_element_size_registry.append(subcontainer.getsize())

    def getbase(self) -> TraceElement:
        if self.parent:
            return self.parent.getloc(self.key)
        return TraceElement(self.memory_space_id, 0, self.getsize(), "")

    def getloc(self, key) -> TraceElement:
        trace = self.getbase()
        if self.is_sparse:
            if key not in self.key_to_loc_map:
                used_offsets = set(self.key_to_loc_map.values())
                remaining_choices = set(range(self.max_len)) - used_offsets
                self.key_to_loc_map[key] = random.choice(
                    tuple(remaining_choices))
            offset = self.key_to_loc_map[key]
        else:
            if key >= self.max_len:
                raise Exception(
                    "This should probably have been a sparse array.")
            offset = key * self.element_size

        return trace._replace(offset=trace.offset + offset,
                              size=self.element_size)


class TrackedList(TrackedContainer, list):
    def __init__(self, *args, is_host_init=False, **kwargs):
        super().__init__(*args, is_host_init=is_host_init, **kwargs)
        # Avoid writes as these could be expensive...
        # for index, value in enumerate(list.__iter__(self)):
        #     # register all values.
        #     self[index] = value
        # self._head = self[0]

    def __iter__(self):
        return TrackedIterator(list.__iter__(self), self)

    def __setitem__(self, key, value, deps: typing.List[int] = None):
        return super(TrackedList, self).__setitem__(key, value, deps)


class TrackedDict(TrackedContainer, dict, is_sparse=True):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for key, value in super().items():
            # register all values.
            self[key] = value

    def update(self, *args, **kwargs):
        for other in args:
            for k, v in other.items():
                self[k] = v

        for k, v in kwargs:
            self[k] = v

    def keys(self):
        return TrackedIterator(super().keys(), self)

    def values(self):
        return TrackedIterator(super().values(), self)

    def items(self):
        return TrackedIterator(super().items(), self)

    def __contains__(self, item):
        TraceRegistry.append(
            lambda: self.getloc(item)._replace(type=TraceElement.READ))
        return super().__contains__(item)


def stub():
    networkx.Graph.node_dict_factory = TrackedDict
    networkx.Graph.node_attr_dict_factory = TrackedDict
    networkx.Graph.adjlist_inner_dict_factory = TrackedDict
    networkx.Graph.adjlist_outer_dict_factory = TrackedDict
    networkx.Graph.edge_attr_dict_factory = TrackedDict



import setuptools

setuptools.setup(
    name="stubbed",
    version="0.1.1",
    packages=setuptools.find_packages(),
    author='Stanford Plasticine Project',
    license='MIT',
    description='A library for obtaining runtime traces',
    install_requires=[
        'numpy',
        'scipy',
        'networkx'
    ]
)



from importlib.abc import SourceLoader, MetaPathFinder
from importlib.util import spec_from_file_location
from importlib.machinery import FileFinder, SourceFileLoader

import importlib

import os
import sys

class HookedLoader(SourceFileLoader):
    def __init__(self, fullname, path):
        super().__init__(fullname, path)
        self.fullname = fullname
        self.path = path

    def get_data(self, path):
        print("Get Data:", path)
        return super().get_data(path)


def install():
    loader_details = HookedLoader, [".py"]
    # insert the path hook ahead of other path hooks
    sys.path_hooks.insert(0, FileFinder.path_hook(loader_details))
    # clear any loaders that might already be in use by the FileFinder
    sys.path_importer_cache.clear()
    importlib.invalidate_caches()

install()